/*
 * Copyright 2019 Google Inc. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.google.turbine.binder.sym;

import com.google.errorprone.annotations.Immutable;
import org.jspecify.nullness.Nullable;

/** A package symbol. */
@Immutable
public class PackageSymbol implements Symbol {

  final String binaryName;

  public PackageSymbol(String binaryName) {
    this.binaryName = binaryName;
  }

  @Override
  public int hashCode() {
    return binaryName.hashCode();
  }

  @Override
  public boolean equals(@Nullable Object obj) {
    return obj instanceof PackageSymbol && binaryName.equals(((PackageSymbol) obj).binaryName);
  }

  @Override
  public String toString() {
    return binaryName.replace('/', '.');
  }

  @Override
  public Kind symKind() {
    return Kind.PACKAGE;
  }

  public String binaryName() {
    return binaryName;
  }
}
