/*
 * Copyright 2016 Google Inc. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.google.turbine.binder.bound;

import com.google.turbine.model.Const;
import com.google.turbine.type.Type;
import java.util.Objects;
import org.jspecify.nullness.Nullable;

/** A class literal constant. */
public class TurbineClassValue extends Const {

  private final Type type;

  public TurbineClassValue(Type type) {
    this.type = type;
  }

  @Override
  public String toString() {
    return String.format("%s.class", type);
  }

  @Override
  public Kind kind() {
    return Kind.CLASS_LITERAL;
  }

  /** The class name. */
  public Type type() {
    return type;
  }

  @Override
  public int hashCode() {
    return Objects.hashCode(type);
  }

  @Override
  public boolean equals(@Nullable Object obj) {
    return obj instanceof TurbineClassValue && type().equals(((TurbineClassValue) obj).type());
  }
}
