/*
 * Copyright 2016 Google Inc. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.google.turbine.binder.bound;

import com.google.common.collect.ImmutableList;
import com.google.common.collect.ImmutableMap;
import com.google.turbine.binder.sym.ClassSymbol;
import com.google.turbine.binder.sym.TyVarSymbol;
import org.jspecify.nullness.Nullable;

/** A bound node that augments {@link BoundClass} with superclasses and interfaces. */
public interface HeaderBoundClass extends BoundClass {
  /** The superclass of the declaration. */
  @Nullable
  ClassSymbol superclass();

  /** The interfaces of the declaration. */
  ImmutableList<ClassSymbol> interfaces();

  /** Declared type parameters. */
  ImmutableMap<String, TyVarSymbol> typeParameters();
}
