/*
 * Copyright 2016 Google Inc. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.google.turbine.binder.lookup;

import static java.util.Objects.requireNonNull;

import com.google.common.collect.ImmutableList;
import com.google.errorprone.annotations.Immutable;
import com.google.turbine.binder.sym.ClassSymbol;
import com.google.turbine.binder.sym.Symbol;
import com.google.turbine.tree.Tree;

/**
 * The result of a canonical type name lookup.
 *
 * <p>For example, a lookup of {@code java.util.HashMap.Entry} may result in the {@link ClassSymbol}
 * for {@code java.util.HashMap}, and the remaining nested type name {@code Entry}.
 */
@Immutable
public class LookupResult {

  /** A top-level class symbol. */
  public Symbol sym() {
    return sym;
  }

  /** The remaining nested type names. */
  public ImmutableList<Tree.Ident> remaining() {
    return remaining;
  }

  private final Symbol sym;
  private final ImmutableList<Tree.Ident> remaining;

  public LookupResult(Symbol sym, LookupKey remaining) {
    this.sym = requireNonNull(sym);
    this.remaining = remaining.hasNext() ? remaining.rest().simpleNames() : ImmutableList.of();
  }
}
