/*
 * Copyright 2016 Google Inc. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.google.turbine.binder.bound;

import com.google.common.collect.ImmutableMap;
import com.google.turbine.binder.sym.ClassSymbol;
import com.google.turbine.model.Const;
import com.google.turbine.type.AnnoInfo;
import org.jspecify.nullness.Nullable;

/** An annotation literal constant. */
public class TurbineAnnotationValue extends Const {

  private final AnnoInfo info;

  public TurbineAnnotationValue(AnnoInfo info) {
    this.info = info;
  }

  @Override
  public String toString() {
    return info.toString();
  }

  @Override
  public Kind kind() {
    return Kind.ANNOTATION;
  }

  /** The annotation declaration. */
  public ClassSymbol sym() {
    return info.sym();
  }

  /** The annotation literal's element-value pairs. */
  public ImmutableMap<String, Const> values() {
    return info.values();
  }

  @Override
  public int hashCode() {
    return info.hashCode();
  }

  @Override
  public boolean equals(@Nullable Object obj) {
    return obj instanceof TurbineAnnotationValue
        && info().equals(((TurbineAnnotationValue) obj).info());
  }

  public AnnoInfo info() {
    return info;
  }
}
